import os

from edge.admin.ui_types import BaseCreator
from edge.path_helpers import *

import edge.pipeline_helpers as helpers
from os.path import join as join_path


def lbs_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules, avaliable_ui_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set(str, str)
    :type used_edge_modules: set(str, str)
    :type avaliable_ui_modules: set of UiIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    # speed
    if ui_entry_point in [Pipeline.Speed.Ui.PIPELINE]:
        creator_ctor = LbsCreator.create

    # mv
    if ui_entry_point in [Pipeline.Mv.Ui.PIPELINE]:
        creator_ctor = LbsCreator.create

    # rtsp
    if ui_entry_point in [Pipeline.Rtsp.Ui.PIPELINE]:
        creator_ctor = LbsCreator.create

    return creator_ctor


class LbsCreator(BaseCreator):
    def __init__(self, id_setupstore):
        """
        :type id_setupstore: str, dict, dict, str
        """
        BaseCreator.__init__(self)

        self._id_setupstore = id_setupstore

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: LbsCreator
        """
        # speed
        if ui_entry_point in [Pipeline.Speed.Ui.PIPELINE]:
            id_setupstore = LbsCreator._serialize(ui_pid, ui_entry_point)
            creator = LbsCreator(id_setupstore)

        # mv
        elif ui_entry_point in [Pipeline.Mv.Ui.PIPELINE]:
            id_setupstore = LbsCreator._serialize(ui_pid, ui_entry_point)
            creator = LbsCreator(id_setupstore)

        # rtsp
        elif ui_entry_point in [Pipeline.Rtsp.Ui.PIPELINE]:
            id_setupstore = LbsCreator._serialize(ui_pid, ui_entry_point)
            creator = LbsCreator(id_setupstore)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: str, dict, dict, str
        """

        if ui_entry_point.ptype == PTYPE_SPEED:
            lbs_edge_identity = Pipeline.Speed.Edge.LBSAPP_0

        elif ui_entry_point.ptype == PTYPE_LPR_RTSP:
            lbs_edge_identity = Pipeline.Rtsp.Edge.LBSAPP_0

        elif ui_entry_point.ptype == PTYPE_LPR_MV:
            lbs_edge_identity = Pipeline.Mv.Edge.LBSAPP_0

        else:
            raise NotImplemented

        # create setupstore.conf
        edge_group = helpers.get_edge_group(ui_pid)
        edge_id = helpers.get_edge_id(edge_group, lbs_edge_identity)
        setupstore_conf_text = get_setupstore_conf_text(edge_id)

        return edge_id, setupstore_conf_text

    def serialize(self):
        edge_id, setupstore_conf_text = self._id_setupstore

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_LBSAPP, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        setupstore_conf_filepath = join_path(edge_jnode_dir, 'setupstore.conf')
        with open(setupstore_conf_filepath, 'w') as text_file:
            text_file.write(setupstore_conf_text)


class LbsDeleter:
    def __init__(self, edge_lbs_dirpath):
        """
        :type edge_lbs_dirpath: str
        """
        self._edge_lbs_dirpath = edge_lbs_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: LbsDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Rtsp.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Rtsp.Edge.LBSAPP_0)
            edge_type = Pipeline.Rtsp.Edge.LBSAPP_0.type
            deleter = LbsDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        if ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Mv.Edge.LBSAPP_0)
            edge_type = Pipeline.Mv.Edge.LBSAPP_0.type
            deleter = LbsDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.LBSAPP_0)
            edge_type = Pipeline.Speed.Edge.LBSAPP_0.type
            deleter = LbsDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_lbs_dirpath)
