from edge import eventlog
from edge.admin.ui_types import *
from edge.path_helpers import *
import edge.pipeline_helpers as helpers
from os.path import join as join_path


def tl_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules, avaliable_ui_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :type avaliable_ui_modules: set of UiIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    ui_modules_set = set()
    # speed
    if ui_entry_point.type in [UI_TYPE_RED_LIGHT, UI_TYPE_PIPELINE]:
        for ui_module in avaliable_ui_modules:
            if ui_module.type in [UI_TYPE_RED_LIGHT]:
                ui_modules_set.add(ui_module)

        if ui_entry_point.type != UI_TYPE_PIPELINE:
            used_edge_modules.add(EdgeIdentity(EDGE_TYPE_TRAFFIC_LIGHT, '0'))
        used_ui_modules.update(ui_modules_set)
        creator_ctor = TlCreator.create
    else:
        return None

    # if len(ui_modules_set) == 0:
    #     # there is no red light detector in this pipeline
    #     return None

    return lambda arg0, arg1, arg2: creator_ctor(arg0, arg1, arg2, ui_modules_set)


class TlCreator(BaseCreator):
    def __init__(self, id_main):
        """
        :type id_main: str, dict
        """
        self._id_main = id_main
        BaseCreator.__init__(self)

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map, avaliable_ui_modules):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :type avaliable_ui_modules: set of UiIdentity
        :rtype: RaCreator
        """
        if ui_entry_point.type in [UI_TYPE_RED_LIGHT, UI_TYPE_PIPELINE]:

            id_main = TlCreator._serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map, avaliable_ui_modules)
            creator = TlCreator(id_main)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map, avaliable_ui_modules):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict | None
        :type avaliable_ui_modules: set of UiIdentity
        :rtype: str, dict, str
        """
        edge_identity = EdgeIdentity(EDGE_TYPE_TRAFFIC_LIGHT, '0')
        edge_group = helpers.get_edge_group(ui_pid)
        edge_id = helpers.get_edge_id(edge_group, edge_identity)
        ui_redlight_rules_id_to_jnode_list = []

        for ui_module in avaliable_ui_modules:
            if ui_module.type == UI_TYPE_RED_LIGHT:
                ui_redlight_rules_id_to_jnode_list.append((ui_module.id, ui_identity_to_jnode_map[ui_module]))

        edge_gid = edge_identity.gid
        spool_dir = get_spool_dir(ui_pid)

        # create main.conf
        main_conf_jnode = {
            # "device": {
            #     "vendor": "komkon",
            #     "params": {
            #         "endpoint": "/dev/ttyUSB3",
            #         "controller": [
            #             {
            #                 "id": 1,
            #                 "red-chan": 1,
            #                 "amber-chan": 2
            #             }
            #         ]
            #     }
            # },
            "changelog": {
                "entry-count": 1200,
                "entry-tout": 125,
                "entry-age": 300
            },
            "track-event": [
                join_path(spool_dir, 'tl%s-sc%s-track' % (edge_gid, edge_gid))
            ]
        }
        # get device params
        import ujson
        if len(ui_redlight_rules_id_to_jnode_list) == 1:
            ui_id, red_light_jnode = ui_redlight_rules_id_to_jnode_list[0]

            main_conf_jnode["delay-usec"] = red_light_jnode.get("delay-usec", 0)

            eventlog.log_d(ujson.dumps(red_light_jnode))
            vendor = red_light_jnode['ro'].get('vendor', 'zonechecker')
            if vendor == 'zonechecker':
                tl_jnode = red_light_jnode['rw']['args']['scene']['area']['traffic-lights']
                params_jnode = {
                    "controller": [
                        {
                            "red-chan": 1,
                            "red-zone": tl_jnode['red'],
                            "amber-chan": 2,
                            "amber-zone": tl_jnode['amber'],
                            "green-chan": 3,
                            "green-zone": tl_jnode['green'],
                            "id": 1
                        }
                    ],
                    "endpoint": join_path(spool_dir, 'mg%s-zc%s-scene' % (edge_gid, edge_gid))
                }

            else:
                params_jnode = red_light_jnode['ro'].get('params')

            if None not in [vendor, params_jnode]:
                main_conf_jnode['device'] = {
                    'vendor': vendor,
                    'params': params_jnode
                }

        # create setupstore.conf
        if len(ui_redlight_rules_id_to_jnode_list) == 1:
            setupstore_conf_text = get_setupstore_conf_text(edge_id, active=True)
        else:
            setupstore_conf_text = get_setupstore_conf_text(edge_id, active=False)

        return edge_id, main_conf_jnode, setupstore_conf_text

    def serialize(self):
        edge_id, main_conf_jnode, setupstore_conf_text = self._id_main

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_TRAFFIC_LIGHT, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_filepath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_filepath, 'w') as json_file:
            ujson.dump(main_conf_jnode, json_file, indent=2)

        setupstore_conf_filepath = join_path(edge_jnode_dir, 'setupstore.conf')
        with open(setupstore_conf_filepath, 'w') as text_file:
            text_file.write(setupstore_conf_text)


class TlDeleter:
    def __init__(self, edge_im_dirpath):
        """
        :type edge_im_dirpath: str
        """
        self._edge_im_dirpath = edge_im_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: RaDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Mv.Edge.TRAFFIC_LIGHT_0)
            edge_type = Pipeline.Mv.Edge.TRAFFIC_LIGHT_0.type
            deleter = TlDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.TRAFFIC_LIGHT_0)
            edge_type = Pipeline.Speed.Edge.TRAFFIC_LIGHT_0.type
            deleter = TlDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Rtsp.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Rtsp.Edge.TRAFFIC_LIGHT_0)
            edge_type = Pipeline.Rtsp.Edge.TRAFFIC_LIGHT_0.type
            deleter = TlDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_im_dirpath)
