import fcntl
import os

from edge.admin.ui_types import LockError

from edge.path_helpers import *
from os.path import join as join_path


def try_lock(lock_dirpath, lock_name):
    if not os.path.exists(lock_dirpath):
        os.makedirs(lock_dirpath)

    lock_filepath = join_path(lock_dirpath, lock_name + '.lock')
    lock_fd = os.open(lock_filepath, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
    try:
        # Request exclusive (EX) non-blocking (NB) advisory lock.
        fcntl.lockf(lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        return lock_fd

    except IOError as x:
        eventlog.log_d("fail: %s" % x.message)
        raise LockError


def unlock(lock_fd):
    try:
        # Request exclusive (EX) non-blocking (NB) advisory lock.
        fcntl.lockf(lock_fd, fcntl.LOCK_UN)
        os.close(lock_fd)
    except IOError as x:
        eventlog.log_d("fail: %s" % x.message)


class LockUiJnodeSet:
    _LOCK_DIRPATH = '/var/run/edge/setupstore/admin/actions/'
    _LOCK_FILENAME_FORMAT = 'pipeline-%s-%s-%s'

    def __init__(self, ui_pid, ui_identity_set):
        """
        :type ui_pid: str
        :type ui_identity_set: set of UiIdentity
        """
        lock_filepath_list = map(lambda _ui_identity: LockUiJnodeSet._create_lock_path(ui_pid, _ui_identity),
                                 ui_identity_set)

        self._lock_fd_list = []
        try:
            for lock_filepath in lock_filepath_list:
                self._lock_fd_list.append(LockUiJnodeSet._lock(lock_filepath))
        except LockError:
            LockUiJnodeSet.release(self)
            raise

    @staticmethod
    def _create_lock_path(ui_pid, ui_identity):
        """
        :type ui_pid: str
        :type ui_identity: UiIdentity
        :rtype: int
        """
        lock_file = LockUiJnodeSet._LOCK_FILENAME_FORMAT % (ui_pid, ui_identity.type, ui_identity.id)
        lock_filepath = join_path(LockUiJnodeSet._LOCK_DIRPATH, lock_file)

        return lock_filepath

    @staticmethod
    def _lock(lock_filepath):
        """
        :type lock_filepath: str
        :rtype: int
        """
        lock_dirpath = os.path.dirname(lock_filepath)
        if not os.path.exists(lock_dirpath):
            os.makedirs(lock_dirpath)

        lock_fd = os.open(lock_filepath, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
        try:
            # Request exclusive (EX) non-blocking (NB) advisory lock.
            fcntl.lockf(lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            return lock_fd

        except IOError as x:
            os.close(lock_fd)
            eventlog.log_d("fail: %s" % x.message)
            raise LockError

    @staticmethod
    def _unlock(lock_fd):
        # Request exclusive (EX) non-blocking (NB) advisory lock.
        fcntl.lockf(lock_fd, fcntl.LOCK_UN)
        os.close(lock_fd)

    def release(self):
        lock_fd_list = []
        for lock_fd in self._lock_fd_list:
            # noinspection PyBroadException
            try:
                LockUiJnodeSet._unlock(lock_fd)
            except Exception:
                eventlog.log_x()
                lock_fd_list.append(lock_fd)

        self._lock_fd_list = lock_fd_list
        fd_count = len(self._lock_fd_list)
        if fd_count:
            eventlog.log_d('fail: unlock %s lock-files' % fd_count)
