import os
import jsonschema
import ujson

from edge import eventlog, path_helpers as ph, pipeline_helpers as oh
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.com import fs

from psuctl import run_action, Endpoint, PsuCtlError

import edge_creator as creator
from ui_pipeline_set import get_pipeline_type


class UiLens:
    def __init__(self, handler):
        """
        :type handler: () -> dict
        """
        self._handler = handler

    @staticmethod
    def _handle_get(ui_pid, entry_point):
        """
        :type ui_pid: str
        :type entry_point: ph.UiIdentity
        """
        videsource_status = oh.get_videsource_status(ui_pid, entry_point.id)
        if videsource_status is None:
            eventlog.log_d('fail: get videosource status')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503,
                error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        elif videsource_status != 0:
            eventlog.log_d('fail: videosource status error')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503,
                error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
        old_jnode = fs.load_json2(ui_jnode_filepath)

        with Endpoint() as endpoint:
            min_iris, max_iris = endpoint.run_action('rotate-iris', {'action': 'get-limits'})
            min_focus, max_focus = endpoint.run_action('rotate-focus', {'action': 'get-limits'})

        if None not in [max_iris, max_focus]:
            old_jnode['rw']['limits'] = {'iris': max_iris, 'focus': max_focus}

        return rh.reply_done(old_jnode['rw'])

    @staticmethod
    def _handle_put(ui_pid, entry_point, new_ui_jnode):
        """
        :type ui_pid: str
        :type entry_point: ph.UiIdentity
        :type new_ui_jnode: dict
        :rtype: dict
        """
        videsource_status = oh.get_videsource_status(ui_pid, entry_point.id)
        if videsource_status is None:
            eventlog.log_d('fail: get videosource status')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503,
                error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        elif videsource_status != 0:
            eventlog.log_d('fail: videosource status error')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503,
                error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        limits_jnode = {}

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
                return False

            # rotate
            new_iris = new_ui_jnode['iris']
            new_focus = new_ui_jnode['focus']

            old_iris = ui_identity_to_jnode_map[entry_point]['rw']['iris']
            old_focus = ui_identity_to_jnode_map[entry_point]['rw']['focus']

            # change jnode
            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode

            with Endpoint() as endpoint:
                min_iris, max_iris = endpoint.run_action('rotate-iris', {'action': 'get-limits'})
                min_focus, max_focus = endpoint.run_action('rotate-focus', {'action': 'get-limits'})
                limits_jnode = {'iris': max_iris, 'focus': max_focus}

                if None in [max_focus, max_iris]:
                    eventlog.log_e('fail: get limits')
                    return None

                if old_iris != new_iris:
                    if new_iris == 0:
                        iris_steps = -(max_iris - min_iris)
                    else:
                        iris_steps = int(new_iris - old_iris)

                    try:
                        endpoint.run_action('rotate-iris', {'steps': iris_steps})
                    except PsuCtlError:
                        eventlog.log_e('fail: rotate iris')
                        return None

                    except fs.LockError:
                        eventlog.log_e('fail: rotate iris (conflict)')
                        raise ui.LockError

                if old_focus != new_focus:
                    if new_focus == 0:
                        focus_steps = -(max_focus - min_focus)
                    else:
                        focus_steps = int(new_focus - old_focus)

                    try:
                        endpoint.run_action('rotate-focus', {'steps': focus_steps})
                    except PsuCtlError:
                        eventlog.log_e('fail: rotate focus')
                        return None

                    except fs.LockError:
                        eventlog.log_e('fail: rotate focus (conflict)')
                        raise ui.LockError

            return True

        try:
            creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)
            if len(limits_jnode):
                new_ui_jnode['limits'] = limits_jnode
            return rh.reply_done(new_ui_jnode)

        except ui.LockError:
            return rh.reply_fail(rh.REPLY_CONFLICT_409)

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLens
        """
        ui_pid = query_args.get_ui_pid()
        ui_type = query_args.get_ui_type()
        ui_id = query_args.get_ui_id()
        if None in [ui_pid, ui_type, ui_id]:
            return UiLens(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiLens(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        if query_args.query_string or query_args.raw_body:
            eventlog.log_d('fail: bad-request')
            return UiLens(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        return UiLens(lambda: UiLens._handle_get(ui_pid, entry_point))

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLens
        """
        ui_pid = query_args.get_ui_pid()
        ui_type = query_args.get_ui_type()
        ui_id = query_args.get_ui_id()
        if None in [ui_pid, ui_type, ui_id]:
            return UiLens(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiLens(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        if query_args.query_string:
            eventlog.log_d('fail: bad-request')
            return UiLens(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load new-ui-jnode
        try:
            new_ui_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None
        except (IOError, ValueError):
            eventlog.log_d('fail: load new-ui-jnode')
            return UiLens(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load ui-schema-jnode
        try:
            ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_d("fail: load ui-schema-jnode (%s)" % x.message)
            return UiLens(lambda: rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500))

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_d("fail: validate new-ui-jnode (%s)" % x.message)
            return UiLens(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # all arguments are valid
        return UiLens(lambda: UiLens._handle_put(ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLens
        """
        if query_args.http_method == 'GET':
            return UiLens._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiLens._deserialize_put(query_args)

        else:
            eventlog.log_d("fail: method-not-allowed")
            return UiLens(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
