import jsonschema
import ujson

from shapely.geometry import polygon

from edge import eventlog, path_helpers as ph, pipeline_helpers as oh
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.com import fs

import edge_creator as creator
from ui_pipeline_set import get_pipeline_type


def _handle_get(ui_pid, entry_point):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    :rtype: dict
    """

    videsource_status = oh.get_videsource_status(ui_pid, entry_point.id)
    if videsource_status is None:
        eventlog.log_d('fail: get videosource status')
        return rh.reply_fail2(
            rh.REPLY_SERVICE_UNAVAILABLE_503,
            error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
            error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

    elif videsource_status != 0:
        eventlog.log_d('fail: videosource status error')
        return rh.reply_fail2(
            rh.REPLY_SERVICE_UNAVAILABLE_503,
            error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
            error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

    old_jnode_file = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
    old_jnode = fs.load_json2(old_jnode_file)

    wh = oh.get_videosource_image_wh(ui_pid=ui_pid, ui_id=entry_point.id)
    if wh is not None:
        w, h = wh
        old_jnode['rw']['stats'] = {
            'scene-wh': [w, h]
        }

    return rh.reply_done(old_jnode['rw'])


def _handle_put(ui_pid, entry_point, new_ui_jnode):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    :type new_ui_jnode: dict
    :rtype: dict
    """

    # check status of videosource
    videsource_status = oh.get_videsource_status(ui_pid, entry_point.id)
    if videsource_status is None:
        eventlog.log_d('fail: get videosource status')
        return rh.reply_fail2(
            rh.REPLY_SERVICE_UNAVAILABLE_503,
            error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
            error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

    elif videsource_status != 0:
        eventlog.log_d('fail: videosource status error')
        return rh.reply_fail2(
            rh.REPLY_SERVICE_UNAVAILABLE_503,
            error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
            error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

    def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
        """
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: bool
        """
        if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
            return False

        ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
        return True

    try:
        creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)

        wh = oh.get_videosource_image_wh(ui_pid=ui_pid, ui_id=entry_point.id)
        if wh is not None:
            w, h = wh
            new_ui_jnode['stats'] = {
                'scene-wh': [w, h]
            }

        return rh.reply_done(new_ui_jnode)

    except ui.LockError:
        return rh.reply_fail(rh.REPLY_CONFLICT_409)


class UiLpr:
    def __init__(self, handler):
        """
        :type handler: () -> dict
        """
        self._handler = handler

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLpr
        """
        #
        http_method = query_args.http_method
        url_query_string = query_args.query_string
        body = ujson.loads(query_args.raw_body) if query_args.raw_body else None
        try:
            ui_pid = query_args.get_ui_pid()
            ui_type = query_args.get_ui_type()
            ui_id = query_args.get_ui_id()
        except IndexError:
            eventlog.log_d('fail: path-info')
            return UiLpr(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiLpr(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            return UiLpr(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        #
        if http_method == 'GET':
            if not body:
                return UiLpr(lambda: _handle_get(ui_pid, entry_point))

        elif http_method == 'PUT':
            schema_jnode_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            request_schema_jnode = fs.load_json2(schema_jnode_filepath)
            if body:
                try:
                    jsonschema.validate(body, request_schema_jnode)
                except (ValueError, jsonschema.ValidationError):
                    return UiLpr(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

                azone_jnode = body['plate-recognizer']['image-azone']
                try:
                    # validate self-intersection of include-mas
                    include_jnode = azone_jnode['include-mask']

                    polymask = polygon.Polygon(include_jnode)
                    if not polymask.is_valid:
                        eventlog.log_d("fail: bad-request (invalid include-mask)")
                        return UiLpr(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

                except (ValueError, IndexError):
                    eventlog.log_d("include-mask contains less then 3 vertexes. using whole image as include-mask")
                    azone_jnode['include-mask'] = []

                try:
                    # validate self-intersection of exclude-mask
                    exclude_jnode = azone_jnode['exclude-mask']

                    for exclude_mask in exclude_jnode:
                        polymask = polygon.Polygon(exclude_mask)
                        if not polymask.is_valid:
                            eventlog.log_d("fail: bad-request (invalid exclude-mask)")
                            return UiLpr(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

                except (ValueError, IndexError):
                    eventlog.log_d("at least one of poly-regions in exclude-mask contains less then 3 vertexes."
                                   " ignoring exclude-mask")
                    azone_jnode['exclude-mask'] = []

                return UiLpr(lambda: _handle_put(ui_pid, entry_point, body))

        else:
            eventlog.log_e('Can not discovered http method')
            return UiLpr(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

        eventlog.log_e('handler_env has been failed')
        return UiLpr(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
